# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

from econpizza import __version__

# -- Project information -----------------------------------------------------

project = 'econpizza'
copyright = '2022, Gregor Boehl'
author = 'Gregor Boehl'

# The full version, including alpha/beta/rc tags
version = __version__
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# extensions = [
# "sphinx.ext.autodoc",
# ]
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates", "**.ipynb_checkpoints"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

autoclass_content = "both"
autodoc_member_order = "groupwise"
latex_use_parts = False


def setup(app):
    app.add_css_file('css/custom.css')


master_doc = 'index'

html_theme_options = {
    "description": " Solve nonlinear heterogeneous agents models",
    "github_button": True,
    "github_repo": "econpizza",
    "github_user": "gboehl",
    "sidebar_collapse": False,
}



#!/bin/python
# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
import numpy as np
import scipy.sparse as ssp


def get_stst_jacobian(model, horizon):
    """Get the steady state Jacobian. Only meant to be used internally.
    """

    pars = jnp.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()
    stst = jnp.array(list(model['stst'].values()))
    nvars = len(model["variables"])

    if model.get('distributions'):
        vfSS = model['steady_state']['decisions']
        distSS = jnp.array(model['steady_state']['distributions'])
        decisions_outputSS = jnp.array(
            model['steady_state']['decisions_output'])

    func_eqns = model['context']["func_eqns"]
    func_backw = model['context'].get('func_backw')
    func_dist = model['context'].get('func_dist')
    zshock = jnp.zeros(len(shocks))

    # TODO: getting the partial Jacobians should be a subfunction
    @jax.jit
    def func_eqns_ssj(x, distributions, decisions_output):
        distributions = distributions.reshape(*distSS.shape, 1)
        decisions_output = decisions_output.reshape(
            *decisions_outputSS.shape, 1)
        return func_eqns(x[:nvars, jnp.newaxis], x[nvars:-nvars, jnp.newaxis], x[-nvars:, jnp.newaxis], stst, zshock, pars, distributions, decisions_output)

    @jax.jit
    def func_backw_ssj(x, vf):
        vf = vf.reshape(*vfSS.shape)
        vf, decisions_output = func_backw(
            x[:nvars, jnp.newaxis], x[nvars:-nvars, jnp.newaxis], x[-nvars:, jnp.newaxis], stst, vf, zshock, pars)[:2]
        return vf.ravel(), decisions_output.ravel()

    @jax.jit
    def func_dist_ssj(distributions, decisions_output):
        distributions = distributions.reshape(*distSS.shape)
        decisions_output = decisions_output.reshape(*decisions_outputSS.shape)
        return func_dist(distributions, decisions_output)[0].ravel()

    if model.get('distributions'):
        # TODO: could still save some time by only cracking those jacobians that we care about
        jac_dist = jax.jacfwd(func_dist_ssj, argnums=(0, 1))(
            distSS.ravel(), decisions_outputSS.ravel())
        jac_f = jax.jacrev(func_eqns_ssj, argnums=(0, 1, 2))(
            jnp.hstack([stst]*3), distSS.ravel(), decisions_outputSS.ravel())
        jac_vf, jac_do = jax.jacfwd(func_backw_ssj, argnums=(0, 1))(
            jnp.hstack([stst]*3), vfSS.ravel())

        # let all this be sparse
        jac_dist2dist = ssp.csr_array(jac_dist[0])
        jac_dist2do = ssp.csc_array(jac_dist[1])
        jac_f2x = ssp.csc_array(jac_f[0])
        jac_f2dist = ssp.csr_array(jac_f[1])
        jac_f2do = ssp.csc_array(jac_f[2])
        (jac_vf2x, jac_vf2vf), (jac_do2x, jac_do2vf) = [
            ssp.csc_array(j) for j in jac_vf], [ssp.csc_array(j) for j in jac_do]

    else:
        jac_f2x = ssp.csc_array(jax.jacrev(lambda x: func_eqns(
            x[:nvars, jnp.newaxis], x[nvars:-nvars, jnp.newaxis], x[-nvars:, jnp.newaxis], stst, zshock, pars))(jnp.hstack([stst]*3)))

    # TODO: calculating the complete Jacobian should be a subfunction
    jac = ssp.lil_array(((horizon-1)*nvars, (horizon+1)*nvars))
    if model.get('distributions'):
        jac_f2x = jac_f2x + jac_f2do.dot(jac_do2x)

    dummy = ssp.lil_array((nvars, (horizon*2+1)*nvars))

    # dummy is centered at t=0 in nvars*horizon => (horizon-1) is t=-1 block
    dummy[:, (horizon-1)*nvars:(horizon+2)*nvars] = jac_f2x

    if model.get('distributions'):
        runner = jac_f2do.dot(jac_do2vf)
        for i in range(1, horizon):
            dummy[:, (horizon+i-1)*nvars:(horizon+i+2)*nvars] = dummy[:,
                                                                      (horizon+i-1)*nvars:(horizon+i+2)*nvars] + runner.dot(jac_vf2x)
            runner = runner.dot(jac_vf2vf)

        runner = jac_f2dist
        for i in range(1, horizon):
            dummy[:, (horizon-i-1)*nvars:(horizon-i+2)*nvars] = dummy[:, (horizon-i-1)
                                                                      * nvars:(horizon-i+2)*nvars] + runner.dot(jac_dist2do).dot(jac_do2x)
            runner = runner.dot(jac_dist2dist)

    for i in range(1, horizon):
        jac[(i-1)*nvars:i*nvars] = dummy[:,
                                         (horizon-i)*nvars:(2*horizon-i+1)*nvars]

    if not model.get('distributions'):
        return jac.tocsr()

    jac_via_dists = ssp.lil_array(((horizon-1)*nvars, (horizon+1)*nvars))
    W = ssp.lil_array((distSS.size, (horizon+1)*nvars))

    runner = jac_dist2do.dot(jac_do2vf)

    for i in range(horizon-1):
        W[:, nvars*i:nvars*(i+3)] = W[:, nvars*i:nvars*(i+3)
                                      ].tocsc() + runner.dot(jac_vf2x)
        runner = runner.dot(jac_vf2vf)

    for i in range(horizon-1):
        jac_via_dists[nvars*i:nvars*(i+1)] = jac_f2dist.dot(W)
        W = jac_dist2dist.dot(W)

    for i in range(horizon-2):
        jac_via_dists[nvars*(i+1):nvars*(i+2), nvars:] = jac_via_dists[nvars*(
            i+1):nvars*(i+2), nvars:] + jac_via_dists[nvars*i:nvars*(i+1), :-nvars]

    return jac.tocsr() + jac_via_dists.tocsr()


def find_path_linear(model, x0, shocks=None, horizon=300, verbose=True):
    """Find the linear expected trajectory given an initial state.

    Parameters
    ----------
    model : dict
        model dict or PizzaModel instance
    x0 : array
        initial state
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`. NOTE: Not (yet) implemented.
    horizon : int, optional
        number of periods until the system is assumed to be back in the steady state. A good idea to set this corresponding to the respective problem. A too large value may be computationally expensive. A too small value may generate inaccurate results
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent

    Returns
    -------
    x : array
        array of the trajectory
    flag : bool
        for consistency. Always returns `True`
    """

    if shocks is not None:
        raise NotImplementedError(
            "Shocks are not (yet) implemented for the linear solution.")

    st = time.time()
    stst = jnp.array(list(model['stst'].values()))
    nvars = len(model["variables"])

    if model['stst_jacobian'] is None:
        stst_jacobian = get_stst_jacobian(model, horizon)
        model['stst_jacobian'] = stst_jacobian
        mess = 'Steady state Jacobian calculated.'
    else:
        stst_jacobian = model['stst_jacobian']
        mess = 'Steady state Jacobian loaded.'

    x0 = jnp.array(list(x0)) - stst
    x = - \
        ssp.linalg.spsolve(
            stst_jacobian[:, nvars:-nvars], stst_jacobian[:, :nvars] @ x0)
    x = jnp.hstack((x0, x)).reshape(-1, nvars) + stst

    if verbose:
        duration = time.time() - st
        print(
            f"(find_path_linear:) {mess} Linear solution done after {duration:1.3f} seconds.")

    return x, True



#!/bin/python
# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
from grgrlib.jaxed import newton_jax, jacfwd_and_val
from .parser.build_functions import get_func_stst_raw


# use a solver that can deal with ill-conditioned jacobians
def solver(jval, fval):
    """A default solver to solve indetermined problems.
    """
    return jnp.linalg.pinv(jval) @ fval


def solve_stst(model, tol=1e-8, tol_newton=None, maxit_newton=30, tol_backwards=None, maxit_backwards=2000, tol_forwards=None, maxit_forwards=5000, force=False, verbose=True, **newton_kwargs):
    """Solves for the steady state.

    Parameters
    ----------
    tol_newton : float, optional
        tolerance of the Newton method, defaults to 1e-8
    maxit_newton : int, optional
        maximum of iterations for the Newton method, defaults to 30
    tol_backwards : float, optional
        tolerance required for backward iteration. Defaults to tol_newton
    maxit_backwards : int, optional
        maximum of iterations for the backward iteration. Defaults to maxit_newton
    tol_forwards : float, optional
        tolerance required for forward iteration. Defaults to tol_newton
    maxit_forwards : int, optional
        maximum of iterations for the forward iteration. Defaults to maxit_newton
    force : bool, optional
        force recalculation of steady state, even if it is already evaluated. Defaults to False
    verbose : bool, optional
        level of verbosity. Defaults to True
    newton_kwargs : keyword arguments
        keyword arguments passed on to the Newton method

    Returns
    -------
    res : dict
        results dictionary from the Newton method
    """

    st = time.time()

    evars = model["variables"]
    func_pre_stst = model['context']["func_pre_stst"]
    par = model["parameters"]
    shocks = model.get("shocks") or ()

    tol_newton = tol if tol_newton is None else tol_newton
    tol_backwards = tol if tol_backwards is None else tol_backwards
    tol_forwards = 1e-2*tol if tol_forwards is None else tol_forwards

    # check if steady state was already calculated
    try:
        cond0 = jnp.allclose(model["stst_used_pars"], jnp.array(
            list(model['steady_state']['fixed_evalued'].values())))
        cond1 = model["stst_used_setup"] == (
            model.get('functions_file_plain'), tol_newton, maxit_newton, tol_backwards, maxit_backwards, tol_forwards, maxit_forwards)
        if cond0 and cond1 and not force:
            if verbose:
                print(
                    f"(solve_stst:) Steady state already {'known' if model['stst_used_success'] else 'failed'}.")

            return model['stst_used_res']
    except KeyError:
        pass

    # reset for recalculation
    model['stst_jacobian'] = None
    model['stacked_func'] = None

    # get all necessary functions
    func_eqns = model['context']['func_eqns']
    func_backw = model['context'].get('func_backw')
    func_stst_dist = model['context'].get('func_stst_dist')

    # get initial values for heterogenous agents
    decisions_output_init = model['init_run'].get('decisions_output')
    exog_grid_vars_init = model['init_run'].get('exog_grid_vars')
    init_vf = model.get('init_vf')

    # get the actual steady state function
    func_stst_raw = get_func_stst_raw(func_pre_stst, func_backw, func_stst_dist, func_eqns, shocks, init_vf, decisions_output_init,
                                      exog_grid_vars_init, tol_backw=tol_backwards, maxit_backw=maxit_backwards, tol_forw=tol_forwards, maxit_forw=maxit_forwards)

    # define jitted stst function that returns jacobian and func. value
    func_stst = jacfwd_and_val(jax.jit(func_stst_raw))
    # store functions
    model["context"]['func_stst_raw'] = func_stst_raw
    model["context"]['func_stst'] = func_stst

    # actual root finding
    res = newton_jax(func_stst, jnp.array(list(model['init'].values())), None, maxit_newton, tol_newton, sparse=False,
                     func_returns_jac=True, solver=solver, verbose=verbose, **newton_kwargs)

    # exchange those values that are identified via stst_equations
    stst_vals, par_vals = func_pre_stst(res['x'])

    model["stst"] = dict(zip(evars, stst_vals))
    model["parameters"] = dict(zip(par, par_vals))
    model["stst_used_pars"] = jnp.array(
        list(model['steady_state']['fixed_evalued'].values()))
    model["stst_used_setup"] = model.get(
        'functions_file_plain'), tol_newton, maxit_newton, tol_backwards, maxit_backwards, tol_forwards, maxit_forwards
    model["stst_used_res"] = res
    model["stst_used_success"] = res['success']

    mess = ''

    if model.get('distributions'):
        # TODO: loosing some time here
        res_backw, res_forw = func_stst_raw(res['x'], True)
        vfSS, decisions_output, exog_grid_vars, cnt_backwards = res_backw
        distSS, cnt_forwards = res_forw
        if jnp.isnan(jnp.array(vfSS)).any() or jnp.isnan(jnp.array(decisions_output)).any():
            mess += f"Backward iteration returns 'NaN's. "
        elif jnp.isnan(distSS).any():
            mess += f"Forward iteration returns 'NaN's. "
        elif distSS.min() < 0:
            mess += f"Distribution contains negative values. "
        if cnt_backwards == maxit_backwards:
            mess += f'Maximum of {maxit_backwards} backwards calls reached. '
        if cnt_forwards == maxit_forwards:
            mess += f'Maximum of {maxit_forwards} forwards calls reached. '
        # TODO: this should loop over the objects in distSS/vfSS and store under the name of the distribution/decisions (i.e. 'D' or 'Va')
        model['steady_state']["distributions"] = distSS
        model['steady_state']['decisions'] = vfSS
        model['steady_state']['decisions_output'] = decisions_output

    # calculate error
    err = jnp.abs(res['fun']).max()

    if err > tol_newton or not res['success']:
        jac = res['jac']
        rank = jnp.linalg.matrix_rank(jac)
        if rank:
            nvars = len(evars)+len(par)
            nfixed = len(model['steady_state']['fixed_evalued'])
            mess += f"Jacobian has rank {rank} for {nvars - nfixed} degrees of freedom ({nvars} variables/parameters, {nfixed} fixed). "
        if not res["success"]:
            mess = f"Steady state FAILED (error is {err:1.2e}). {res['message']} {mess}"
        else:
            mess = f"{res['message']} WARNING: Steady state error is {err:1.2e}. {mess}"
    elif verbose:
        duration = time.time() - st
        mess = f"Steady state found in {duration:1.5g} seconds. {res['message']}" + (
            ' WARNING: ' + mess if mess else '')

    if mess:
        print(f"(solve_stst:) {mess}")

    return res




"""Build dynamic functions."""

import jax
import jax.numpy as jnp
from scipy import sparse
from grgrlib.jaxed import jacfwd_and_val, jacrev_and_val


def get_func_stst_raw(func_pre_stst, func_backw, func_stst_dist, func_eqns, shocks, init_vf, decisions_output_init, exog_grid_vars_init, tol_backw, maxit_backw, tol_forw, maxit_forw):
    """Get a function that evaluates the steady state
    """

    def cond_func(cont):
        (vf, _, _), vf_old, cnt = cont
        cond0 = jnp.abs(vf - vf_old).max() > tol_backw
        cond1 = cnt < maxit_backw
        return cond0 & cond1

    def body_func_raw(cont, x, par):
        (vf, _, _), _, cnt = cont
        return func_backw(x, x, x, x, vf, [], par), vf, cnt + 1

    def func_backw_ext(x, par):

        def body_func(cont): return body_func_raw(cont, x, par)

        (vf, decisions_output, exog_grid_vars), _, cnt = jax.lax.while_loop(
            cond_func, body_func, ((init_vf, decisions_output_init, exog_grid_vars_init), init_vf+1, 0))

        return vf, decisions_output, exog_grid_vars, cnt

    def func_stst_raw(y, full_output=False):

        x, par = func_pre_stst(y)
        x = x[..., None]

        if not func_stst_dist:
            return func_eqns(x, x, x, x, jnp.zeros(len(shocks)), par)

        vf, decisions_output, exog_grid_vars, cnt_backw = func_backw_ext(
            x, par)
        dist, cnt_forw = func_stst_dist(decisions_output, tol_forw, maxit_forw)

        # TODO: for more than one dist this should be a loop...
        decisions_output_array = decisions_output[..., None]
        dist_array = dist[..., None]

        if full_output:
            return (vf, decisions_output, exog_grid_vars, cnt_backw), (dist, cnt_forw)

        return func_eqns(x, x, x, x, [], par, dist_array, decisions_output_array)

    return func_stst_raw


def get_stacked_func_dist(pars, func_backw, func_dist, func_eqns, x0, stst, vfSS, distSS, zshock, tshock, horizon, nvars, endpoint, has_distributions, shock):
    """Get a function that returns the (flattend) value and Jacobian of the stacked aggregate model equations.
    """

    nshpe = (nvars, horizon-1)

    def backwards_step(carry, i):

        vf, X = carry
        vf, decisions_output, exog_grid_vars = func_backw(
            X[:, i], X[:, i+1], X[:, i+2], stst, vf, [], pars)

        return (vf, X), decisions_output

    def forwards_step(carry, i):

        dist_old, decisions_output_storage = carry
        dist = func_dist(dist_old, decisions_output_storage[..., i])

        return (dist, decisions_output_storage), dist_old

    from grgrlib.jaxed import jax_print

    def stacked_func_dist(x, full_output=False):

        X = jnp.hstack((x0, x, endpoint)).reshape(horizon+1, -1).T

        if has_distributions:
            # backwards step
            _, decisions_output_storage = jax.lax.scan(
                backwards_step, (vfSS, X), jnp.arange(horizon-1), reverse=True)
            decisions_output_storage = jnp.moveaxis(
                decisions_output_storage, 0, -1)
            # forwards step
            _, dists_storage = jax.lax.scan(
                forwards_step, (distSS, decisions_output_storage), jnp.arange(horizon-1))
            dists_storage = jnp.moveaxis(dists_storage, 0, -1)
        else:
            decisions_output_storage, dists_storage = [], []

        if full_output:
            return decisions_output_storage, dists_storage

        out = func_eqns(X[:, :-2].reshape(nshpe), X[:, 1:-1].reshape(nshpe), X[:, 2:].reshape(
            nshpe), stst, zshock, pars, dists_storage, decisions_output_storage)

        return out

    return stacked_func_dist


def get_stacked_func(pars, func_eqns, stst, x0, horizon, nvars, endpoint, zshock, tshock, shock, dist_dummy=[], decisions_dummy=[]):
    """Get a function that returns the (flattend) value and Jacobian of the stacked aggregate model equations IGNORING the effects of distributions.
    """

    jac_vmap = jax.vmap(jacfwd_and_val(lambda x: func_eqns(
        x[:nvars, None], x[nvars:-nvars, None], x[-nvars:, None], stst, zshock, pars, dist_dummy, decisions_dummy)))
    jac_shock = jacfwd_and_val(lambda x: func_eqns(
        x[:nvars], x[nvars:-nvars], x[-nvars:], stst, tshock, pars, dist_dummy, decisions_dummy))

    def stacked_func(x):

        X = jnp.vstack((x0, x.reshape((horizon - 1, nvars)), endpoint))
        Y = jnp.hstack((X[:-2], X[1:-1], X[2:]))
        out, jac_parts = jac_vmap(Y)

        # the ordering is ((period1(f=1,...,N), ..., periodT(f=1,...,N)) x (period1(var=1,...,N), ..., periodT(var=1,...,N)))
        J = sparse.lil_array(((horizon-1)*nvars, (horizon-1)*nvars))

        if shock is None:
            J[:nvars, :nvars * 2] = jac_parts[0, :, nvars:]
        else:
            out_shocked, jac_part_shocked = jac_shock(X[:3].ravel())
            J[:nvars, :nvars * 2] = jac_part_shocked[:, nvars:]
            out = out.at[0].set(out_shocked)
        J[-nvars:, (horizon-3) * nvars:horizon *
          nvars] = jac_parts[horizon-2, :, :-nvars]

        for t in range(1, horizon-2):
            J[nvars*t:nvars*(t+1), (t-1)*nvars:(t+2)*nvars] = jac_parts[t]

        return out.ravel(), J

    return stacked_func


def get_combined_funcs(stacked_func_dist, stacked_func, mask_out, use_jacrev):
    """Combine Jacobians and values from stacked_func and stacked_func_dist
    """

    if use_jacrev:
        jav_stacked_func_dist = jacrev_and_val(
            jax.jit(lambda x: stacked_func_dist(x)[mask_out]))
    else:
        jav_stacked_func_dist = jacfwd_and_val(
            jax.jit(lambda x: stacked_func_dist(x)[mask_out]))

    def combined_funcs(x):

        fval, jacval = stacked_func(x)
        fval_dists, jacval_dists = jav_stacked_func_dist(x)
        fval = fval.at[mask_out].set(fval_dists)
        jacval[mask_out] = jacval_dists

        return fval, jacval

    return combined_funcs



"""Functions that write other functions.
"""


def compile_func_basics_str(evars, par, shocks):

    func_str = f"""
        \n ({"".join(v + "Lag, " for v in evars)}) = XLag
        \n ({"".join(v + ", " for v in evars)}) = X
        \n ({"".join(v + "Prime, " for v in evars)}) = XPrime
        \n ({"".join(v + "SS, " for v in evars)}) = XSS
        \n ({"".join(p + ", " for p in par)}) = pars
        \n ({"".join(s + ", " for s in shocks)}) = shocks"""

    return func_str


def compile_backw_func_str(evars, par, shocks, inputs, outputs, calls, exog_grid_var_names):
    """Compile all information to a string that defines the backward function for 'decisions'.
    """

    if isinstance(calls, str):
        calls = calls.splitlines()

    func_str = f"""def func_backw(XLag, X, XPrime, XSS, VFPrime, shocks, pars):
            {compile_func_basics_str(evars, par, shocks)}
            \n ({"".join(v + ", " for v in inputs)}) = VFPrime
            \n %s
            \n return jnp.array(({"".join(v[:-5] + ", " for v in inputs)})), jnp.array(({", ".join(v for v in outputs)})), ({', '.join(v for v in exog_grid_var_names)})
            """ % '\n '.join(calls)

    return func_str


def compile_stst_func_str(evars, par, stst, init):
    """Compile all information from 'equations' section' to a string that defines the function.
    """

    stst_str = '; '.join([f'{v} = {stst[v]}' for v in stst])

    # compile the final function string
    func_pre_stst_str = f"""def func_pre_stst(Y):
        \n ({"".join(v + ", " for v in init)}) = Y
        \n {stst_str}
        \n X = ({"".join(v + ", " for v in evars)})
        \n par = ({"".join(p + ", " for p in par)})
        \n return jnp.array(X), jnp.array(par)"""

    return func_pre_stst_str


def compile_func_dist_str(distributions, decisions_outputs):
    """Compiles the transition functions for distributions.
    """

    if len(distributions) > 1:
        raise NotImplementedError(
            'More than one distribution is not yet implemented.')

    # already prepare for more than one distributions
    # each distribution gets an own string
    func_stst_dist_str_tpl = ()
    func_dist_str_tpl = ()

    for dist_name in distributions:

        dist = distributions[dist_name]
        exog = [v for v in dist if dist[v]['type'] in (
            'exogenous', 'custom_exogenous', 'time_varying_exogenous')]
        endo = [v for v in dist if dist[v]['type'] == 'endogenous']

        if len(exog) > 1:
            raise NotImplementedError(
                'Exogenous distributions with more than one dimension are not yet implemented.')

        func_dist_str_tpl = f"\n endog_inds0, endog_probs0 = interp.interpolate_coord_robust({dist[endo[0]]['grid_variables']}, {endo[0]})",

        if len(endo) == 1:
            func_stst_dist_str_tpl = func_dist_str_tpl + \
                (f"\n {dist_name}, {dist_name}_cnt = dists.stationary_distribution_forward_policy_1d(endog_inds0, endog_probs0, {dist[exog[0]]['grid_variables'][2]}, tol, maxit)",)
            func_dist_str_tpl += f"\n {dist_name} = {dist[exog[0]]['grid_variables'][2]}.T @ dists.forward_policy_1d({dist_name}, endog_inds0, endog_probs0)",

        elif len(endo) == 2:
            func_dist_str_tpl += f"\n endog_inds1, endog_probs1 = interp.interpolate_coord_robust({dist[endo[1]]['grid_variables']}, {endo[1]})",
            func_stst_dist_str_tpl = func_dist_str_tpl + \
                (f"\n {dist_name}, {dist_name}_cnt = dists.stationary_distribution_forward_policy_2d(endog_inds0, endog_inds1, endog_probs0, endog_probs1, {dist[exog[0]]['grid_variables'][2]}, tol, maxit)",)
            func_dist_str_tpl += f"""
                \n forwarded_dist = dists.forward_policy_2d({dist_name}, endog_inds0, endog_inds1, endog_probs0, endog_probs1)
                \n {dist_name} = expect_transition({dist[exog[0]]['grid_variables'][2]}.T, forwarded_dist)
                """,

        else:
            raise NotImplementedError(
                'Endogenous distributions with more than two dimension are not yet implemented.')

    # join the tuples to strings that define the final functions
    func_stst_dist_str = f"""def func_stst_dist(decisions_outputs, tol, maxit):
        \n ({", ".join(decisions_outputs)},) = decisions_outputs
        \n {"".join(func_stst_dist_str_tpl)}
        \n max_cnt = jnp.max({"".join(d + '_cnt, ' for d in distributions.keys())})
        \n return jnp.array(({"".join(d + ', ' for d in distributions.keys())})), max_cnt"""

    func_dist_str = f"""def func_dist(distributions, decisions_outputs):
        \n ({", ".join(decisions_outputs)},) = decisions_outputs
        \n ({"".join(d+', ' for d in distributions)}) = distributions
        \n {"".join(func_dist_str_tpl)}
        \n return jnp.array(({"".join(d + ', ' for d in distributions.keys())}))"""

    return func_stst_dist_str, func_dist_str


def compile_eqn_func_str(evars, eqns, par, eqns_aux, shocks, distributions, decisions_outputs):
    """Compile all information from 'equations' section' to a string that defines the function.
    """

    # start compiling root_container
    for i, eqn in enumerate(eqns):
        if "=" in eqn:
            lhs, rhs = eqn.split("=")
            eqns[i] = f"root_container{i} = {lhs} - ({rhs})"
        else:
            eqns[i] = f"root_container{i} = {eqn}"

    if isinstance(eqns_aux, str):
        eqns_aux = eqns_aux.splitlines()

    eqns_aux_stack = "\n ".join(eqns_aux) if eqns_aux else ""
    eqns_stack = "\n ".join(eqns)

    # compile the final function string
    func_str = f"""def func_eqns(XLag, X, XPrime, XSS, shocks, pars, distributions=[], decisions_outputs=[]):
        {compile_func_basics_str(evars, par, shocks)}
        \n ({"".join(d+', ' for d in distributions)}) = distributions
        \n ({"".join(d+', ' for d in decisions_outputs)}) = decisions_outputs
        \n {eqns_aux_stack}
        \n {eqns_stack}
        \n {"return jnp.array([" + ", ".join("root_container"+str(i) for i in range(len(evars))) + "]).T.ravel()"}"""

    return func_str



#!/bin/python
# -*- coding: utf-8 -*-

import os
import jax
import time
import jax.numpy as jnp
import scipy.sparse as ssp
from grgrlib.jaxed import newton_jax, jacfwd_and_val, jacrev_and_val
from ..parser.build_functions import *


def find_path_stacking(
    model,
    x0=None,
    shock=None,
    horizon=250,
    init_guess=None,
    endpoint=None,
    tol=None,
    maxit=None,
    use_jacrev=True,
    verbose=True,
    **solver_kwargs,
):
    """Find the expected trajectory given an initial state.

    Parameters
    ----------
    model : dict
        model dict or PizzaModel instance
    x0 : array
        initial state
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`
    horizon : int, optional
        number of periods until the system is assumed to be back in the steady state. A good idea to set this corresponding to the respective problem. A too large value may be computationally expensive. A too small value may generate inaccurate results
    init_guess : array, optional
        a first guess on the trajectory. Not necessary in general
    endpoint : array, optional
        the final state at `horizon`. Defaults to the steay state if `None`
    tol : float, optional
        convergence criterion. Defaults to 1e-8
    maxit : int, optional
        number of iterations. Default is 30.
    use_jacrev : bool, optional
        whether to use reverse mode or forward mode automatic differentiation. By construction, reverse AD is faster, but does not work for all types of functions. Defaults to True
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent
    solver_kwargs : optional
        any additional keyword arguments will be passed on to the solver

    Returns
    -------
    x : array
        array of the trajectory
    flag : bool
        returns True if the solver was successful, else False
    """

    st = time.time()

    stst = jnp.array(list(model["stst"].values()))
    nvars = len(model["variables"])
    pars = jnp.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()
    # load functions
    func_eqns = model['context']["func_eqns"]
    func_backw = model['context'].get('func_backw')
    func_dist = model['context'].get('func_dist')

    tol = 1e-8 if tol is None else tol
    maxit = 30 if maxit is None else maxit

    x0 = jnp.array(list(x0)) if x0 is not None else stst
    x_init = jnp.ones((horizon + 1, nvars)) * stst
    x_init = x_init.at[0].set(x0)

    if init_guess is not None:
        x_init[1: len(init_guess)] = init_guess[1:]

    zshock = jnp.zeros(len(shocks))
    tshock = jnp.copy(zshock)
    if shock is not None:
        tshock = tshock.at[shocks.index(shock[0])].set(shock[1])
        if model.get('distributions'):
            print("(find_stack:) Warning: shocks for heterogenous agent models are not yet fully supported. Use adjusted steady state values as x0 instead.")

    endpoint = endpoint if endpoint is not None else stst

    if model.get('distributions'):
        vfSS = model['steady_state']['decisions']
        distSS = jnp.array(model['steady_state']['distributions'])
        decisions_outputSS = jnp.array(
            model['steady_state']['decisions_output'])

        # get values of func_eqns that are independent of the distribution
        stst_in_t = stst[:, None]

        def func_eqns_from_dist(dist, decisions_output): return func_eqns(
            stst_in_t, stst_in_t, stst_in_t, stst, zshock, pars, dist[..., None], decisions_output[..., None])

        _, mask_out_jvp = jax.jvp(func_eqns_from_dist, (distSS, decisions_outputSS), (
            jnp.ones_like(distSS), jnp.ones_like(decisions_outputSS)))
        mask_out = jnp.tile(mask_out_jvp.astype(bool), horizon-1)

        # compile dist-specific functions
        stacked_func_dist_raw = get_stacked_func_dist(pars, func_backw, func_dist, func_eqns, x0, stst,
                                                      vfSS, distSS, zshock, tshock, horizon, nvars, endpoint, model.get('distributions'), shock)
        stacked_func_dist = jax.jit(
            stacked_func_dist_raw, static_argnames='full_output')
        model['context']['stacked_func_dist'] = stacked_func_dist
        dist_dummy = distSS[..., None]
        decisions_output_dummy = decisions_outputSS[..., None]
    else:
        dist_dummy = decisions_output_dummy = []

    stacked_func = get_stacked_func(pars, func_eqns, stst, x0, horizon, nvars,
                                    endpoint, zshock, tshock, shock, dist_dummy, decisions_output_dummy)

    if verbose:
        print("(find_stack:) Solving stack (size: %s)..." %
              (horizon*nvars))

    if model.get('distributions'):
        stacked_func = get_combined_funcs(
            stacked_func_dist, stacked_func, mask_out, use_jacrev)
    res = newton_jax(stacked_func, x_init[1:-1].flatten(), None, maxit, tol,
                     sparse=True, func_returns_jac=True, verbose=verbose, **solver_kwargs)

    # calculate error
    err = jnp.abs(res['fun']).max()
    x = x_init.at[1:-1].set(res['x'].reshape((horizon - 1, nvars)))

    mess = res['message']

    if err > tol or not res['success']:
        mess += f" Max. error is {err:1.2e}."
        verbose = True

    if verbose:
        duration = time.time() - st
        sucess = 'done' if res['success'] else 'FAILED'
        print(
            f"(find_path:) Stacking {sucess} after {duration:1.3f} seconds. " + mess)

    return x, not res['success']



#!/bin/python
# -*- coding: utf-8 -*-

import jax
import numpy as np
from scipy.linalg import block_diag
from grgrlib import klein, speed_kills


def solve_linear_state_space(
    model,
    x=None,
    eps=1e-5,
    raise_error=True,
    check_contraction=False,
    lti_max_iter=1000,
    verbose=True,
):
    """Does half-way SGU, solves the model using linear time iteration and uses Klein's method to check for determinancy if the solution fails"""

    if model.get('distributions'):
        raise Exception(
            "A linear state-space solution for models with distributions is not implemented.")

    evars = model["variables"]
    func = model['context']["func_eqns"]
    par = np.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()
    stst = list(model["stst"].values())
    nshc = len(shocks)
    nsts = len(stst)

    if x is None:
        x = np.array(stst)

    xmult = x.copy()
    xmult[np.isclose(x, 0)] = 1

    AA = np.empty((nsts, nsts))
    BB = AA.copy()
    CC = AA.copy()
    DD = np.empty((nsts, nshc))

    zshock = np.zeros(len(shocks))
    fx = func(x, x, x, x, np.zeros(nshc), par)

    # use numdifftools if possible
    AA = jax.jacfwd(lambda err: func(x, x, err, x, zshock, par))(x) * xmult
    BB = jax.jacfwd(lambda err: func(x, err, x, x, zshock, par))(x) * xmult
    CC = jax.jacfwd(lambda err: func(err, x, x, x, zshock, par))(x) * xmult
    DD = jax.jacfwd(lambda err: func(x, x, x, x, err, par))(zshock)
    DD = DD.reshape((nsts, len(shocks)))

    A = np.pad(AA, ((0, nshc), (0, nshc)))
    B = block_diag(BB, np.eye(nshc))
    C = np.block([[CC, DD], [np.zeros((nshc, A.shape[1]))]])

    model["ABC"] = A, B, C

    I = np.eye(nsts + nshc)
    Z = np.zeros_like(I)
    P = np.block([[B, A], [I, Z]])
    M = np.block([[C, Z], [Z, -I]])

    mess = ""
    success = True

    try:
        try:
            lam = -speed_kills(
                P, M, nsts + nshc, max_iter=lti_max_iter, verbose=verbose - 1
            )[1]

        except:
            _, lam = klein(P, M, nsts + nshc, verbose=verbose - 1)

        model["lin_pol"] = -lam[:nsts, :nsts], -lam[:nsts, nsts:]
        mess = "All eigenvalues are good"

    except Exception as error:
        success = False
        if raise_error:
            raise error
        else:
            mess = str(error).strip()
            if mess[-1] == ".":
                mess = mess[:-1]

    if check_contraction:
        A = np.linalg.inv(BB) @ AA
        B = np.linalg.inv(BB) @ CC

        Aev = np.abs(np.linalg.eig(A)[0])
        Aev_err = Aev > 1
        Bev = np.abs(np.linalg.eig(B)[0])
        Bev_err = Bev > 1
        flag += Aev_err.any() or Bev_err.any()
        mess += ", but " if success else ""
        if Aev_err.any():
            mess += "%s forward looking EV%s larger than unity (%s)" % (
                Aev_err.sum(),
                "s are" if Aev_err.sum() > 1 else " is",
                *Aev[Aev_err],
            )
        if Aev_err.any() and Bev_err.any():
            mess += " and "
        if Bev_err.any():
            mess += "%s backward looking EV%s larger than unity (%s)" % (
                Bev_err.sum(),
                "s are" if Bev_err.sum() > 1 else " is",
                *Bev[Bev_err],
            )

    if mess and verbose:
        print(f"(solve_linear:) {mess}{'' if mess[-1] in '.?!' else '.'}")

    return model["ABC"]


def find_path_linear_state_space(model, shock, T, x, use_linear_guess):
    """Solves the expected trajectory given the linear model.
    """

    if model.get("lin_pol") is not None:

        stst = np.array(list(model["stst"].values()))
        sel = np.isclose(stst, 0)

        shocks = model.get("shocks") or ()
        tshock = np.zeros(len(shocks))
        if shock is not None:
            tshock[shocks.index(shock[0])] = shock[1]

        x_lin = np.empty_like(x)
        x_lin[0] = x[0]
        x_lin[0][~sel] = (x[0][~sel] / stst[~sel] - 1)

        for t in range(T):
            x_lin[t + 1] = model["lin_pol"][0] @ x_lin[t]

            if not t:
                x_lin[t + 1] += model["lin_pol"][1] @ tshock

        x_lin[:, ~sel] = ((1 + x_lin) * stst)[:, ~sel]

        if use_linear_guess:
            return x_lin.copy(), x_lin
        else:
            return x, x_lin
    else:
        return x, None



#!/bin/python
# -*- coding: utf-8 -*-
"""Functions for model parsing yaml into a working model instance. Involves a lot of dynamic function definition...
"""

from copy import copy
import yaml
import re
import os
import tempfile
import jax
import jaxlib
import jax.numpy as jnp
from copy import deepcopy
# from jax.numpy import log, exp, sqrt, maximum, minimum # maybe better to make this explicit
from grgrlib import load_as_module
from inspect import getmembers, isfunction
from jax.experimental.host_callback import id_print as jax_print
from .utilities import grids, dists, interp
from .parser.compile_functions import *
from .parser.checks import *

# initialize model cache
cached_mdicts = ()
cached_models = ()


def parse(mfile):
    """parse from yaml file"""

    f = open(mfile)
    mtxt = f.read()
    f.close()

    mtxt = mtxt.replace("^", "**")
    mtxt = re.sub(r"@ ?\n", " ", mtxt)
    # try to detect if `~` wants to be a `-`
    mtxt = mtxt.replace("\n ~ ", "\n - ")
    mtxt = mtxt.replace("\n  ~ ", "\n  - ")
    mtxt = mtxt.replace("   ~ ", "   - ")

    # get dict
    model = yaml.safe_load(mtxt)
    # create nice shortcuts
    model['path'] = mfile
    model["pars"] = model.get("parameters")
    model["vars"] = model["variables"]

    return model


def eval_strs(vdict, context={}):
    """Evaluate a dictionary of strings into a given context
    """

    if vdict is None:
        return None

    for v in vdict:
        if isinstance(vdict[v], str):
            context[v] = eval(vdict[v], context)
            vdict[v] = eval(v, context)
        else:
            context[v] = vdict[v]

    return vdict


def parse_external_functions_file(model):
    """Parse the functions file.
    """

    try:
        # prepare path
        if not os.path.isabs(model["functions_file"]):
            yaml_dir = os.path.dirname(model["path"])
            model["functions_file"] = os.path.join(
                yaml_dir, model["functions_file"])

        # store content
        f = open(model["functions_file"])
        model['functions_file_plain'] = f.read()
        f.close()

    except KeyError:
        pass

    return


def initialize_context():
    context = globals().copy()
    default_funcs = {'log': jnp.log,
                     'exp': jnp.exp,
                     'sqrt': jnp.sqrt,
                     'max': jnp.maximum,
                     'min': jnp.minimum}

    context.update(default_funcs)
    return context


def load_external_functions_file(model, context):
    """Load the functions file as a module.
    """

    try:
        # load as a module
        module = load_as_module(model["functions_file"])

        def func_or_compiled(func): return isinstance(
            func, jaxlib.xla_extension.CompiledFunction) or isfunction(func)
        for m in getmembers(module, func_or_compiled):
            context[m[0]] = m[1]

    except KeyError:
        pass

    return False


def compile_init_values(evars, pars, initvals, stst):
    """Combine all available information in initial guesses.
    """

    ufixed_vars = {v: 1.1 for v in evars if v not in stst}
    ufixed_pars = {v: 1.1 for v in pars if v not in stst}

    # get inital values to test the function
    init = {**ufixed_vars, **ufixed_pars}

    # structure: aggregate values first, then values of decisions functions
    if initvals is not None:
        for v in init:
            # assign aggregate values
            if v in initvals:
                init[v] = initvals[v]

    return init


def define_subdict_if_absent(parent, sub):
    try:
        return parent[sub]
    except KeyError:
        parent[sub] = {}
        return parent[sub]


def define_function(func_str, context):
    """Define functions from string. Writes the function into a temporary file in order to get meaningful debug traces.
    """

    # use a termporary file to get nice debug traces if things go wrong
    tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)
    tmpf.write(func_str)
    tmpf.close()

    # define the function
    exec(compile(open(tmpf.name).read(), tmpf.name, "exec"), context)

    return tmpf.name


def get_exog_grid_var_names(distributions):
    """WIP. So far unused.
    """
    # NOTE: this will be important when implementing that grid parameters are endogenous variables
    # NOTE: when activated, backward calls already return exogenous grid vars (exog_grid_var). They are not yet stacked, and not yet an input to forward calls

    exog_grid_var_names = ()

    if False:
        # if distributions:
        for dist_name in distributions:

            dist = distributions[dist_name]
            for v in dist:
                if dist[v]['type'] in ('exogenous', 'custom_exogenous'):
                    exog_grid_var_names += tuple(dist[v]['grid_variables'])

    return exog_grid_var_names


def load(
    model,
    raise_errors=True,
    verbose=True
):
    """Load model from dict or yaml file.

    Parameters
    ----------
    model : dict or string
        either a dictionary or the path to a yaml file to be parsed
    raise_errors : bool, optional
        whether to raise errors while checking. False will let the model fail siliently for debugging. Defaults to True
    verbose : bool, optional
        inform that parsing is done. Defaults to True

    Returns
    -------
    model : PizzaModel instance
        The parsed model
    """

    from .__init__ import PizzaModel

    global cached_mdicts, cached_models

    if isinstance(model, str):
        full_path = model
        model = parse(model)
        model['path'] = full_path

    model = PizzaModel(model)

    # load file with additional functions as module (if it exists)
    parse_external_functions_file(model)
    mdict_raw = deepcopy(model)

    # check if model is already cached
    if model in cached_mdicts:
        model = cached_models[cached_mdicts.index(model)]
        print("(parse:) Loading cached model.")
        return model

    model['context'] = initialize_context()
    load_external_functions_file(model, model['context'])

    defs = model.get("definitions")
    defs = '' if defs is None else defs
    defs = '\n'.join(defs) if isinstance(defs, list) else defs
    exec(defs, model['context'])

    eqns = model["equations"].copy()

    # check if there are dublicate variables
    evars = check_dublicates_and_determinancy(model["variables"], eqns)
    # check if each variable is defined in time t (only defining xSS does not give a valid root)
    check_if_defined(evars, eqns, model.get('skip_check_if_defined'))

    # create fixed (time invariant) grids
    grids.create_grids(model.get('distributions'), model["context"])

    shocks = model.get("shocks") or ()
    _ = define_subdict_if_absent(model, "steady_state")
    fixed_values = define_subdict_if_absent(
        model["steady_state"], "fixed_values")
    stst = eval_strs(fixed_values, context=model['context'])
    model["steady_state"]['fixed_evalued'] = stst

    par = define_subdict_if_absent(model, "parameters")
    model["parameters"] = par
    model["root_options"] = {}

    # collect number of foward and backward looking variables
    model["no_fwd"] = sum(
        var + "Prime" in "".join(model["equations"]) for var in evars)
    model["no_bwd"] = sum(var + "Lag" in "".join(model["equations"])
                          for var in evars)

    tmpf_names = ()
    # initialize storage for all function strings
    model['func_strings'] = {}

    # NOTE: currently disabled
    exog_grid_var_names = get_exog_grid_var_names(model.get('distributions'))

    # get function strings for decisions and distributions, if they exist
    if model.get('decisions'):
        decisions_outputs = model['decisions']['outputs']
        decisions_inputs = model['decisions']['inputs']
        model['func_strings']["func_backw"] = compile_backw_func_str(
            evars, par, shocks, decisions_inputs, decisions_outputs, model['decisions']['calls'], exog_grid_var_names)
        tmpf_names += define_function(model['func_strings']
                                      ['func_backw'], model['context']),
    else:
        decisions_outputs = []
        decisions_inputs = []

    if model.get('distributions'):
        dist_names = list(model['distributions'].keys())
        func_stst_dist_str, func_dist_str = compile_func_dist_str(
            model['distributions'], decisions_outputs)
        # store both strings
        model['func_strings']["func_stst_dist"] = func_stst_dist_str
        model['func_strings']["func_dist"] = func_dist_str
        # execute them
        tmpf_names += define_function(func_stst_dist_str, model['context']),
        tmpf_names += define_function(func_dist_str, model['context']),
    else:
        dist_names = []

    # collect initial guesses
    init_guesses = eval_strs(model["steady_state"].get(
        "init_guesses"), context=model['context'])
    model["init"] = init = compile_init_values(evars, par, init_guesses, stst)

    # get strings that contains the function definitions
    model['func_strings']["func_pre_stst"] = compile_stst_func_str(
        evars, par, stst, init)
    model['func_strings']["func_eqns"] = compile_eqn_func_str(evars, deepcopy(eqns), par, eqns_aux=model.get(
        'aux_equations'), shocks=shocks, distributions=dist_names, decisions_outputs=decisions_outputs)

    # test if model works. Writing to tempfiles helps to get nice debug traces if not
    tmpf_names += define_function(model['func_strings']
                                  ["func_eqns"], model['context']),
    tmpf_names += define_function(model['func_strings']
                                  ['func_pre_stst'], model['context']),

    # get the initial decision functions
    if model.get('decisions'):
        init_vf_list = [model['steady_state']['init_guesses'][dec_input]
                        for dec_input in model['decisions']['inputs']]  # let us for now assume that this must be present
        model['init_vf'] = jnp.array(init_vf_list)

        # check if initial decision functions and the distribution have same shapes
        check_shapes(model['distributions'], model['init_vf'], dist_names)

    # try if function works on initvals
    model['init_run'] = {}
    try:
        check_initial_values(model, shocks, par)
    except:
        if raise_errors:
            raise

    if verbose:
        print("(load:) Parsing done.")

    # add new model to cache
    cached_mdicts += (mdict_raw,)
    cached_models += (model,)

    return model



#!/bin/python
# -*- coding: utf-8 -*-

import os
from econpizza.__init__ import *
import econpizza as ep

filepath = os.path.dirname(__file__)


def test_bh(create=False):

    mod = load(example_bh, raise_errors=False)
    _ = mod.solve_stst()

    state = np.zeros(len(mod["variables"]))
    state[:-1] = [0.1, 0.2, 0.0]

    x, flag = find_path_shooting(
        mod, state, T=50, max_horizon=500, tol=1e-8, verbose=2)

    path = os.path.join(filepath, "test_storage", "bh.npy")

    if create:
        np.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = np.load(path)

        assert flag == 0
        assert np.allclose(x, test_x)


def test_nk(create=False):

    mod = load(example_nk)
    _ = mod.solve_stst()

    state = mod["stst"].copy()
    state["beta"] *= 1.02

    x, flag = find_path_shooting(mod, state.values(), T=10,
                                 max_horizon=10, verbose=2)

    path = os.path.join(filepath, "test_storage", "nk.npy")

    if create:
        np.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = np.load(path)

        assert flag == 0
        assert np.allclose(x, test_x)


def test_stacked(create=False):

    mod = load(example_nk)
    _ = mod.solve_stst()

    shk = ("e_beta", 0.02)

    x, flag = find_path_stacking(mod, shock=shk)

    path = os.path.join(filepath, "test_storage", "stacked.npy")

    if create:
        np.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = np.load(path)

        assert flag == 0
        assert np.allclose(x, test_x)


def test_hank(create=False):

    mod_dict = ep.parse(example_hank)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-4)

    x0 = mod['stst'].copy()
    x0['beta'] *= 1.01  # setting a shock on the discount factor

    x, flag = mod.find_path(x0.values(), horizon=10)
    x_lin, _ = mod.find_path_linear(x0.values(), horizon=10)
    het_vars = mod.get_het_vars(x)
    dist = het_vars['dist']

    path_x = os.path.join(filepath, "test_storage", "hank.npy")
    path_x_lin = os.path.join(filepath, "test_storage", "hank_lin.npy")
    path_dist = os.path.join(filepath, "test_storage", "hank_dist.npy")

    if create:
        np.save(path_x, x)
        np.save(path_x_lin, x_lin)
        np.save(path_dist, dist)
        print(f'Test file updated at {path_x},{path_x_lin} and {path_dist}')
    else:
        test_x = np.load(path_x)
        test_x_lin = np.load(path_x_lin)
        test_dist = np.load(path_dist)

        assert flag == 0
        assert np.allclose(x, test_x)
        assert np.allclose(x_lin, test_x_lin)
        assert np.allclose(dist, test_dist)


def test_hank_labor(create=False):

    mod_dict = ep.parse(example_hank_labor)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-6)

    x0 = mod['stst'].copy()
    x0['beta'] *= 1.01  # setting a shock on the discount factor

    x, flag = mod.find_path(x0.values(), horizon=10, use_jacrev=False)

    path = os.path.join(filepath, "test_storage", "hank_labor.npy")

    if create:
        np.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = np.load(path)

        assert flag == 0
        assert np.allclose(x, test_x)


def test_hank2(create=False):

    mod_dict = ep.parse(example_hank2)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-6)

    x0 = mod['stst'].copy()
    x0['beta'] *= 1.01  # setting a shock on the discount factor

    x, flag = mod.find_path(x0.values(), horizon=10)

    path = os.path.join(filepath, "test_storage", "hank2.npy")

    if create:
        np.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = np.load(path)

        assert flag == 0
        assert np.allclose(x, test_x)



#!/bin/python
# -*- coding: utf-8 -*-

import logging
import jax
import os
import numpy as np
from .parsing import parse, load
from .steady_state import solve_stst
from .solvers.shooting import find_path_shooting
from .solvers.stacking import find_path_stacking
from .solvers.solve_linear import find_path_linear

__version__ = '0.2.1'

jax.config.update("jax_enable_x64", True)
# set number of cores for XLA
os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={os.cpu_count()}"


class PizzaModel(dict):
    def __init__(self, *args, **kwargs):
        super(PizzaModel, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def get_het_vars(self, xst):
        """Get all disaggregated variables for a given trajectory of aggregate variables.

        Parameters
        ----------

        self : PizzaModel
            the model instance
        xst : array
            a trajectory of aggregate variables

        Returns
        -------
        rdict : dict
            a dictionary of the outputs of the decision and distributions stage
        """

        stacked_func = self['context']['stacked_func_dist']
        decisions_outputs = self['decisions']['outputs']
        dist_names = list(self['distributions'].keys())

        het_vars = stacked_func(xst[1:-1].ravel(), True)

        rdict = {oput: het_vars[0][i]
                 for i, oput in enumerate(decisions_outputs)}
        rdict.update({oput: het_vars[1][i]
                     for i, oput in enumerate(dist_names)})

        return rdict


PizzaModel.find_path = find_path_stacking
PizzaModel.find_path_stacking = find_path_stacking
PizzaModel.find_path_linear = find_path_linear
PizzaModel.find_path_shooting = find_path_shooting
PizzaModel.solve_stst = solve_stst

logging.basicConfig(level=logging.INFO)

pth = os.path.dirname(__file__)

example_nk = os.path.join(pth, "examples", "nk.yaml")
example_bh = os.path.join(pth, "examples", "bh.yaml")
example_dsge = os.path.join(pth, "examples", "med_scale_nk.yaml")
example_hank = os.path.join(pth, "examples", "hank.yaml")
example_hank_labor = os.path.join(pth, "examples", "hank_labor.yaml")
example_hank2 = os.path.join(pth, "examples", "hank2.yaml")




"""Grids and Markov chains"""

import jax
import jax.numpy as jnp
from scipy.stats import norm
from .dists import stationary_distribution


def log_grid(amax, n, amin=0):
    """Create grid between amin and amax that is equidistant in logs."""
    pivot = jnp.abs(amin) + 0.25
    a_grid = jnp.geomspace(amin + pivot, amax + pivot, n) - pivot
    a_grid = a_grid.at[0].set(amin)  # make sure *exactly* equal to amin
    return a_grid


def mean(x, pi):
    """Mean of discretized random variable with support x and probability mass function pi."""
    return jnp.sum(pi * x)


def variance(x, pi):
    """Variance of discretized random variable with support x and probability mass function pi."""
    return jnp.sum(pi * (x - jnp.sum(pi * x)) ** 2)


def markov_rouwenhorst(rho, sigma, N=7):
    """Rouwenhorst method analog to markov_tauchen"""

    # Explicitly typecast N as an integer, since when the grid constructor functions
    # (e.g. the function that makes a_grid) are implemented as blocks, they interpret the integer-valued calibration
    # as a float.
    N = int(N)

    # parametrize Rouwenhorst for n=2
    p = (1 + rho) / 2
    Pi = jnp.array([[p, 1 - p], [1 - p, p]])

    # implement recursion to build from n=3 to n=N
    for n in range(3, N + 1):
        P1, P2, P3, P4 = (jnp.zeros((n, n)) for _ in range(4))
        P1 = P1.at[:-1, :-1].set(p * Pi)
        P2 = P2.at[:-1, 1:].set((1 - p) * Pi)
        P3 = P3.at[1:, :-1].set((1 - p) * Pi)
        P4 = P4.at[1:, 1:].set(p * Pi)
        Pi = P1 + P2 + P3 + P4
        Pi = Pi.at[1:-1].divide(2)

    # invariant distribution and scaling
    pi = stationary_distribution(jnp.array(Pi.T))
    s = jnp.linspace(-1, 1, N)
    s *= (sigma / jnp.sqrt(variance(s, pi)))
    y = jnp.exp(s) / jnp.sum(pi * jnp.exp(s))

    return y, pi, Pi


def create_grids(distributions, context):
    """Get the strings of functions that define the grids.
    """

    if not distributions:
        return

    grid_strings = ()

    # create strings of the function that define the grids
    for dist_name, dist in distributions.items():
        for grid_name, g in dist.items():

            if g['type'] == 'exogenous':
                grid_strings += f"{', '.join(v for v in g['grid_variables'])} = grids.markov_rouwenhorst(rho={g['rho']}, sigma={g['sigma']}, N={g['n']})",

            elif g['type'] == 'custom_exogenous':
                grid_strings += f"{', '.join(v for v in g['grid_variables'])} = {g['call']}",

            elif g['type'] == 'time_varying_exogenous':
                # in this case the grid must be defined in some stage in the yaml
                # TODO: this is half-way implemented (during backward calls)
                raise NotImplementedError

            elif g['type'] == 'endogenous':
                # as above
                if not all([i not in g for i in ['min', 'max', 'n']]):
                    grid_strings += f"{g['grid_variables']} = grids.log_grid(amin={g['min']}, amax={g['max']}, n={g['n']})",

    # execute all of them
    for grid_str in grid_strings:
        exec(grid_str, context)

    return



#!/bin/python
# -*- coding: utf-8 -*-

import jax
import jax.numpy as jnp


def tmat_from_exog(probs, D):

    nZ, nX = D.shape
    tmat = jnp.zeros((nX*nZ, nX*nZ))

    i = jnp.arange(nZ)*nX
    xi = jnp.arange(nX)[:, None]
    zi = jnp.arange(nZ)[:, None, None]
    tmat = tmat.at[xi+zi*nX, xi +
                   i].set(probs.broadcast_in_dim((nZ, nX, nZ), (0, 2)))

    return tmat


def tmat_from_endo(x_i, probs):

    nZ, nX = x_i.shape
    tmat = jnp.zeros((nX*nZ, nX*nZ))

    ix = jnp.arange(nX*nZ)
    j = jnp.arange(nZ).repeat(nX)*nX
    i = x_i.ravel()
    pi = probs.ravel()

    tmat = tmat.at[i+j, ix].add(pi)
    tmat = tmat.at[i+1+j, ix].add(1-pi)

    return tmat


def expect_transition(Pi, X):
    """If Pi is a matrix, multiply Pi times the ith dimension of X and return"""

    shape = X.shape
    # iterate forward using Pi
    X = Pi @ X.reshape((shape[0], -1))
    # reverse steps
    X = X.reshape((Pi.shape[0], *shape[1:]))

    return X


def forward_policy_1d(D, x_i, x_pi):

    nZ, _ = D.shape
    Dnew = jnp.zeros_like(D)

    j = jnp.arange(nZ)[:, None]

    Dnew = Dnew.at[j, x_i].add(D * x_pi)
    Dnew = Dnew.at[j, x_i+1].add(D * (1 - x_pi))

    return Dnew


def stationary_distribution_forward_policy_1d(endog_inds, endog_probs, exog_probs, tol=1e-10, maxit=1000):

    dist = jnp.ones_like(endog_inds, dtype=jnp.float64)
    dist /= dist.sum()

    def cond_func(cont):
        dist, dist_old, cnt = cont
        cond0 = jnp.abs(dist-dist_old).max() > tol
        cond1 = cnt < maxit
        return cond0 & cond1

    def body_func(cont):
        dist, _, cnt = cont
        return exog_probs.T @ forward_policy_1d(dist, endog_inds, endog_probs), dist, cnt + 1

    dist, _, cnt = jax.lax.while_loop(cond_func, body_func, (dist, dist+1, 0))
    return dist, cnt


def forward_policy_2d(D, x_i, y_i, x_pi, y_pi):

    nZ, _, _ = D.shape
    Dnew = jnp.zeros_like(D)

    j = jnp.arange(nZ)[:, None, None]

    Dnew = Dnew.at[j, x_i, y_i].add(y_pi * x_pi * D)
    Dnew = Dnew.at[j, x_i+1, y_i].add(y_pi * (1 - x_pi) * D)
    Dnew = Dnew.at[j, x_i, y_i+1].add((1 - y_pi) * x_pi * D)
    Dnew = Dnew.at[j, x_i+1, y_i+1].add((1 - y_pi) * (1 - x_pi) * D)

    return Dnew


def stationary_distribution_forward_policy_2d(endog_inds0, endog_inds1, endog_probs0, endog_probs1, exog_probs, tol=1e-10, maxit=1000):
    # TODO: can be merged with stationary_distribution_forward_policy_1d

    dist = jnp.ones_like(endog_inds0, dtype=jnp.float64)
    dist /= dist.sum()

    def cond_func(cont):
        dist, dist_old, cnt = cont
        cond0 = jnp.abs(dist-dist_old).max() > tol
        cond1 = cnt < maxit
        return cond0 & cond1

    def body_func(cont):
        dist, _, cnt = cont
        pre_exo_dist = forward_policy_2d(
            dist, endog_inds0, endog_inds1, endog_probs0, endog_probs1)
        new_dist = expect_transition(exog_probs.T, pre_exo_dist)
        return new_dist, dist, cnt + 1

    dist, _, cnt = jax.lax.while_loop(cond_func, body_func, (dist, dist+1, 0))
    return dist, cnt


def stationary_distribution(T):
    """Find invariant distribution of a Markov chain by unit eigenvector.
    NOTE: jax has no autodiff support for eig.
    """

    v, w = jnp.linalg.eig(T)

    # using sorted args instead of np.isclose is neccessary for jax-jitting
    args = jnp.argsort(v)
    unit_ev = w[:, args[-1]]

    return unit_ev.real / unit_ev.real.sum()
